/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * messages.h:
 * This file defines the message types for the
 * trace messages. Also defines the constants for
 * file system operations.
 */

#ifndef __MESSAGES_H
#define __MESSAGES_H

//The below two fields are changed from short to int since the
//size can reach upto 70K
typedef unsigned int msg_len_t; /* length field in trace message */
typedef unsigned int field_len_t; /* length field in argument */
typedef int msg_seqnum_t; /* Sequence Number Field in Trace Message */


typedef unsigned short field_t;   /* field type */
#define FIELD_LENGTH_MASK (1 << (sizeof(field_t) * 8 - 2))
typedef unsigned short op_t;      /* operation type */

#define OP_BASE         0
/* super block ops */
#define OP_READ_INODE            (OP_BASE + 1)               // 1
#define OP_WRITE_INODE           (OP_READ_INODE + 1)         // 2
#define OP_PUT_INODE             (OP_WRITE_INODE + 1)        // 3
#define OP_DELETE_INODE          (OP_PUT_INODE + 1)          // 4
#define OP_PUT_SUPER             (OP_DELETE_INODE + 1)       // 5
#define OP_STATFS                (OP_PUT_SUPER + 1)          // 6
#define OP_REMOUNT_FS            (OP_STATFS + 1)             // 7
#define OP_CLEAR_INODE           (OP_REMOUNT_FS + 1)         // 8
#define OP_UMOUNT_BEGIN          (OP_CLEAR_INODE + 1)        // 9
#define OP_WRITE_SUPER           (OP_UMOUNT_BEGIN + 1)       // 10
/* dentry ops */
#define OP_D_REVALIDATE          (OP_WRITE_SUPER + 1)        // 11
#define OP_D_HASH                (OP_D_REVALIDATE + 1)       // 12
#define OP_D_COMPARE             (OP_D_HASH + 1)             // 13
#define OP_D_DELETE              (OP_D_COMPARE + 1)          // 14
#define OP_D_RELEASE             (OP_D_DELETE + 1)           // 15
#define OP_D_IPUT                (OP_D_RELEASE + 1)          // 16
#define OP_DGET                  (OP_D_IPUT + 1)             // 17
#define OP_DPUT                  (OP_DGET + 1)               // 18
/* inode ops */
#define OP_CREATE                (OP_DPUT + 1)               // 19
#define OP_LOOKUP                (OP_CREATE + 1)             // 20
#define OP_LINK                  (OP_LOOKUP + 1)             // 21
#define OP_UNLINK                (OP_LINK + 1)               // 22
#define OP_SYMLINK               (OP_UNLINK + 1)             // 23
#define OP_MKDIR                 (OP_SYMLINK + 1)            // 24
#define OP_RMDIR                 (OP_MKDIR + 1)              // 25
#define OP_MKNOD                 (OP_RMDIR + 1)              // 26
#define OP_RENAME                (OP_MKNOD + 1)              // 27
#define OP_READLINK              (OP_RENAME + 1)             // 28
#define OP_FOLLOW_LINK           (OP_READLINK + 1)           // 29
#define OP_INODE_REVALIDATE      (OP_FOLLOW_LINK + 1)        // 30
#define OP_SETATTR               (OP_INODE_REVALIDATE + 1)   // 31
#define OP_PERMISSION            (OP_SETATTR + 1)            // 32
/* file ops */
#define OP_LLSEEK                (OP_PERMISSION + 1)         // 33
#define OP_READ                  (OP_LLSEEK + 1)             // 34
#define OP_WRITE                 (OP_READ + 1)               // 35
#define OP_READDIR               (OP_WRITE + 1)              // 36
#define OP_POLL                  (OP_READDIR + 1)            // 37
#define OP_FLUSH                 (OP_POLL + 1)               // 38
#define OP_MMAP                  (OP_FLUSH + 1)              // 39
#define OP_LOCK                  (OP_MMAP + 1)               // 40
#define OP_FSYNC                 (OP_LOCK + 1)               // 41
#define OP_FASYNC                (OP_FSYNC + 1)              // 42
#define OP_IOCTL                 (OP_FASYNC + 1)             // 43
#define OP_OPEN                  (OP_IOCTL + 1)              // 44
#define OP_RELEASE               (OP_OPEN + 1)               // 45
/* mmap ops */
#define OP_READPAGE              (OP_RELEASE + 1)            // 46
#define OP_READPAGES             (OP_READPAGE + 1)           // 47
#define OP_WRITEPAGE             (OP_READPAGES + 1)          // 48
#define OP_WRITEPAGES            (OP_WRITEPAGE + 1)          // 49
#define OP_PREPARE_WRITE         (OP_WRITEPAGES + 1)         // 50
#define OP_COMMIT_WRITE          (OP_PREPARE_WRITE + 1)      // 51
/* memory management */
#define OP_LOAD			 (OP_COMMIT_WRITE + 1)       // 52
#define OP_UNLOAD		 (OP_LOAD + 1)               // 53
/* OP_READ per page data for profiling */
#define OP_READPP		 (OP_UNLOAD + 1)             // 54
/* OPs missing */
#define OP_TRUNCATE		 (OP_READPP + 1)             // 55
#define OP_DIRTY_INODE		 (OP_TRUNCATE + 1)           // 56
#define OP_REVALIDATE		 (OP_DIRTY_INODE + 1)        // 57
#define OP_GETXATTR		 (OP_REVALIDATE + 1)         // 58
#define OP_SETXATTR		 (OP_GETXATTR + 1)           // 59
#define OP_LISTXATTR		 (OP_SETXATTR + 1)           // 60
#define OP_SHOW_OPTIONS		 (OP_LISTXATTR + 1)          // 61
#define OP_BMAP			 (OP_SHOW_OPTIONS + 1)       // 62
#define OP_SYNC_PAGE		 (OP_BMAP + 1)               // 63
#define OP_DIRECT_IO		 (OP_SYNC_PAGE + 1)          // 64
#define OP_FLUSHPAGE		 (OP_DIRECT_IO + 1)          // 65
#define OP_RELEASEPAGE		 (OP_FLUSHPAGE + 1)          // 66
#define OP_FH_TO_DENTRY		 (OP_RELEASEPAGE + 1)        // 67
#define OP_DENTRY_TO_FH		 (OP_FH_TO_DENTRY + 1)       // 68
#define OP_WRITE_SUPER_LOCKFS	 (OP_DENTRY_TO_FH + 1)       // 69
#define OP_UNLOCKFS		 (OP_WRITE_SUPER_LOCKFS + 1) // 70
#define OP_SYNC_FS		 (OP_UNLOCKFS + 1)           // 71
#define OP_CREATE_THREAD         (OP_SYNC_FS + 1)            // 72
#define OP_MAX          	 (OP_CREATE_THREAD + 1)            // 73
/* all pre ops have same value as the OP_ value */
#define POST_OP_BIT (1 << (8 * sizeof(op_t) - 1))
static inline op_t pre_op(op_t op) { return op; }
static inline op_t post_op(op_t op) { return (op |= POST_OP_BIT); }
#define IS_POST_OP(op) ((op) & POST_OP_BIT)
#define GET_OP(op) (IS_POST_OP((op)) ? ((op) & ~POST_OP_BIT) : (op))
/* base value for non file op messages */
#define MSG_META_BASE   (1 << (8 * sizeof(op_t) - 2))
#define MSG_START                (MSG_META_BASE + 0)
#define MSG_STOP                 (MSG_META_BASE + 1)
#define MSG_PNAME                (MSG_META_BASE + 2)
#define MSG_SYSCALL              (MSG_META_BASE + 3)

/* common arguments */
/* all type length value type arguments are encoded with last but one bit
 * set enabled. Since we're using short, this would mean an OR operation with
 * 0x4000. Make sure that the values of arguments are all < 0x4000, that is,
 * with the range of 0 - 16383 inclusive
 */

#define ARG_INVAL	    0
#define ARG_PATHNAME        1
#define ARG_TIMESTAMP       2
#define ARG_UTIMESTAMP      3
#define ARG_INUM            4
#define ARG_INT_FLAGS       5
#define ARG_LOFFSET         6
#define ARG_IATTR           7
#define ARG_REQUESTED_SIZE  8
#define ARG_DATA            9
#define ARG_FD              10
#define ARG_UID             11
#define ARG_GID             12
#define ARG_RET             13
#define ARG_MSG_DESC        14
#define ARG_MSG_ID          15
#define ARG_CHILD_INUM      16
#define ARG_PID             17
#define ARG_FIELD_ID        18
#define ARG_FIELD_LEN       19
#define ARG_FIELD_DESC      20
#define ARG_PNAME           21
// uname(2) fields
#define ARG_SYSNAME         22
#define ARG_RELEASE         23
#define ARG_VERSION         24
#define ARG_MACHINE         25
// for link
#define ARG_NEW_PATHNAME    26
#define ARG_DIR_INUM        27
// system call ID
#define ARG_SYSCALL         28
#define ARG_SYSCALL_NAME    29
// for calls with two inodes, dentries
#define ARG_NEW_INUM        30
// general string
#define ARG_STR             31
// general unsigned long (size_t)
#define ARG_ULONG           32
// for checksum
#define ARG_CKSUM           33
// These are for replaying, will not be in the original trace file
#define ARG_INODE           34
#define ARG_DENTRY          35
#define ARG_FILE            36
#define ARG_VFSMNT          37
// more fields
#define ARG_DENTRY_PTR      38
#define ARG_INODE_PTR       39
#define ARG_FILE_PTR        40
#define ARG_FILE_FLAGS      41
// needed with ops where we need to have more than one object of a type.
#define ARG_NEW_DENTRY_PTR  42
#define ARG_NEW_INODE_PTR   43

// page-based ops
#define ARG_PAGE_INDEX      44
#define ARG_PAGES_INDEX     45
#define ARG_WRITE_FROM      46
#define ARG_WRITE_TO        47
//
#define ARG_UULONG	    48
#define ARG_KTHREAD_DATA    49

/* Non kernel stuff, used by the user level trace parsing library */
//#ifndef __KERNEL__
#define MSG_ENTRY(M) { M, #M }
struct msg_desc_t {
	op_t msg_type;
	char *str_type;
};

struct arg_desc_t {
	field_t         arg_type;
	char*           str_type;
	field_len_t     size;
};

#ifdef NEED_DESCRIPTIVE_MESSAGES

struct msg_desc_t msg_desc[] = {
	MSG_ENTRY(OP_READ_INODE),
	MSG_ENTRY(OP_WRITE_INODE),
	MSG_ENTRY(OP_PUT_INODE),
	MSG_ENTRY(OP_DELETE_INODE),
	MSG_ENTRY(OP_PUT_SUPER),
	MSG_ENTRY(OP_STATFS),
	MSG_ENTRY(OP_REMOUNT_FS),
	MSG_ENTRY(OP_CLEAR_INODE),
	MSG_ENTRY(OP_UMOUNT_BEGIN),
	MSG_ENTRY(OP_WRITE_SUPER),
	/* dentry ops */
	MSG_ENTRY(OP_D_REVALIDATE),
	MSG_ENTRY(OP_D_HASH),
	MSG_ENTRY(OP_D_COMPARE),
	MSG_ENTRY(OP_D_DELETE),
	MSG_ENTRY(OP_D_RELEASE),
	MSG_ENTRY(OP_D_IPUT),
	/* inode ops */
	MSG_ENTRY(OP_CREATE),
	MSG_ENTRY(OP_LOOKUP),
	MSG_ENTRY(OP_LINK),
	MSG_ENTRY(OP_UNLINK),
	MSG_ENTRY(OP_SYMLINK),
	MSG_ENTRY(OP_MKDIR),
	MSG_ENTRY(OP_RMDIR),
	MSG_ENTRY(OP_MKNOD),
	MSG_ENTRY(OP_RENAME),
	MSG_ENTRY(OP_READLINK),
	MSG_ENTRY(OP_FOLLOW_LINK),
	MSG_ENTRY(OP_INODE_REVALIDATE),
	MSG_ENTRY(OP_SETATTR),
	MSG_ENTRY(OP_PERMISSION),
	/* file ops */
	MSG_ENTRY(OP_LLSEEK),
	MSG_ENTRY(OP_READ),
	MSG_ENTRY(OP_WRITE),
	MSG_ENTRY(OP_READDIR),
	MSG_ENTRY(OP_POLL),
	MSG_ENTRY(OP_FLUSH),
	MSG_ENTRY(OP_MMAP),
	MSG_ENTRY(OP_LOCK),
	MSG_ENTRY(OP_FSYNC),
	MSG_ENTRY(OP_FASYNC),
	MSG_ENTRY(OP_IOCTL),
	MSG_ENTRY(OP_OPEN),
	MSG_ENTRY(OP_RELEASE),
	MSG_ENTRY(OP_READPAGE),
	MSG_ENTRY(OP_READPAGES),
	MSG_ENTRY(OP_WRITEPAGE),
	MSG_ENTRY(OP_PREPARE_WRITE),
	MSG_ENTRY(OP_COMMIT_WRITE),
	MSG_ENTRY(OP_LOAD),
	MSG_ENTRY(OP_UNLOAD),
	MSG_ENTRY(OP_TRUNCATE),
	MSG_ENTRY(OP_DIRTY_INODE),
	MSG_ENTRY(OP_REVALIDATE),
	MSG_ENTRY(OP_GETXATTR),
	MSG_ENTRY(OP_SETXATTR),
	MSG_ENTRY(OP_LISTXATTR),
	MSG_ENTRY(OP_SHOW_OPTIONS),
	MSG_ENTRY(OP_BMAP),
	MSG_ENTRY(OP_SYNC_PAGE),
	MSG_ENTRY(OP_DIRECT_IO),
	MSG_ENTRY(OP_FLUSHPAGE),
	MSG_ENTRY(OP_RELEASEPAGE),
	MSG_ENTRY(OP_FH_TO_DENTRY),
	MSG_ENTRY(OP_DENTRY_TO_FH),
	MSG_ENTRY(OP_WRITE_SUPER_LOCKFS),
	MSG_ENTRY(OP_UNLOCKFS),
	MSG_ENTRY(OP_SYNC_FS),
	MSG_ENTRY(MSG_START),
	MSG_ENTRY(MSG_STOP),
	MSG_ENTRY(MSG_PNAME),
	MSG_ENTRY(MSG_SYSCALL),
	{ -1, NULL }
};

/*
 * variable length args are put in this structure with zero length
 * because we don't look into this array for the length field
 */

#define ARG_ENTRY(M, S) { M, #M, S }
struct arg_desc_t arg_desc[] = {
	ARG_ENTRY(ARG_PATHNAME, 0),
	ARG_ENTRY(ARG_TIMESTAMP, sizeof(int)),
	ARG_ENTRY(ARG_UTIMESTAMP, 2 * sizeof(int)),
	ARG_ENTRY(ARG_INUM, sizeof(int)),
	ARG_ENTRY(ARG_INT_FLAGS, sizeof(int)),
	ARG_ENTRY(ARG_LOFFSET, sizeof(loff_t)),
	ARG_ENTRY(ARG_IATTR, 0),
	ARG_ENTRY(ARG_REQUESTED_SIZE, sizeof(int)),
	ARG_ENTRY(ARG_DATA, 0),
	ARG_ENTRY(ARG_FD, sizeof(int)),
	ARG_ENTRY(ARG_UID, sizeof(uid_t)),
	ARG_ENTRY(ARG_GID, sizeof(uid_t)),
	ARG_ENTRY(ARG_RET, sizeof(int)),
	ARG_ENTRY(ARG_MSG_DESC, 0),
	ARG_ENTRY(ARG_MSG_ID, sizeof(op_t)),
        ARG_ENTRY(ARG_CHILD_INUM, sizeof(int)),
	ARG_ENTRY(ARG_PID, sizeof(pid_t)),
	ARG_ENTRY(ARG_FIELD_ID, sizeof(field_t)),
	ARG_ENTRY(ARG_FIELD_LEN, sizeof(field_len_t)),
	ARG_ENTRY(ARG_FIELD_DESC, 0),
	ARG_ENTRY(ARG_PNAME, 0),
	ARG_ENTRY(ARG_SYSNAME, 0),
        ARG_ENTRY(ARG_RELEASE, 0),
	ARG_ENTRY(ARG_VERSION, 0),
	ARG_ENTRY(ARG_MACHINE, 0),
	ARG_ENTRY(ARG_NEW_PATHNAME, 0),
	ARG_ENTRY(ARG_DIR_INUM, sizeof(int)),
	ARG_ENTRY(ARG_SYSCALL, sizeof(unsigned short)),
	ARG_ENTRY(ARG_SYSCALL_NAME, 0),
	ARG_ENTRY(ARG_NEW_INUM, sizeof(int)),
	ARG_ENTRY(ARG_STR, 0),
	ARG_ENTRY(ARG_ULONG, sizeof(unsigned long)),
	ARG_ENTRY(ARG_CKSUM, 0),
	ARG_ENTRY(ARG_DENTRY_PTR, sizeof(int)),
	ARG_ENTRY(ARG_NEW_DENTRY_PTR, sizeof(int)),
	ARG_ENTRY(ARG_FILE_PTR, sizeof(int)),
	ARG_ENTRY(ARG_INODE_PTR, sizeof(int)),
	ARG_ENTRY(ARG_NEW_INODE_PTR, sizeof(int)),
	ARG_ENTRY(ARG_FILE_FLAGS, sizeof(int)),
	ARG_ENTRY(ARG_PAGE_INDEX, sizeof(int)),
	ARG_ENTRY(ARG_PAGES_INDEX, sizeof(int)),
	ARG_ENTRY(ARG_WRITE_FROM, sizeof(int)),
	ARG_ENTRY(ARG_WRITE_TO, sizeof(int)),
	ARG_ENTRY(ARG_UULONG, sizeof(unsigned long long)),
	{ -1, NULL, 0}
};

#else
extern struct msg_desc_t msg_desc[];
extern struct arg_desc_t arg_desc[];

#endif
//#endif /* !__ KERNEL__ */

#endif /* __MESSAGES_H */
