/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#ifndef _REPLAYK_H
#define _REPLAYK_H

#define NO_VERSION

#include "fist.h"
#include <linux/module.h>
#include <linux/sys.h>
#include "replay.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/page-flags.h>
#endif

struct replay {
	struct replay_command* srcmdp;
	struct res_alloc_table* ratp;
	struct file* trace;
	struct file* data;
	unsigned int cmd_num;
	unsigned int res_num;
	unsigned int cur_cmd;
	int b_compare_data;
};

#define CHECK_RES_TYPE(chk_arg, chk_type) do { \
		if (ratp[ chk_arg ].type != chk_type) { \
                        print_error_message("CHECK_RES_TYPE FAILED - expected type = %d, actual type = %d",chk_type,ratp[chk_arg].type); \
			err = -EINVAL; \
			goto out; \
		} \
	} while(0);

int lookup_rat(struct replay* r, int ival, int type);

int replay_open(struct replay_command* rcmd, struct replay* r);
int replay_lookup(struct replay_command* rcmd, struct replay* r);
int replay_flush(struct replay_command* rcmd, struct replay* r);
int replay_release(struct replay_command* rcmd, struct replay* r);
int replay_write(struct replay_command* rcmd, struct replay* r);
int replay_read(struct replay_command* rcmd, struct replay* r);
int replay_create(struct replay_command* rcmd, struct replay* r);
int replay_symlink(struct replay_command* srcmdp,struct replay* r);
int replay_link(struct replay_command* srcmdp,struct replay* r);
int replay_rename(struct replay_command* srcmdp,struct replay* r);
int replay_prefetch_load(struct replay_command* rcmd, struct replay* r);
int replay_prefetch_unload(struct replay_command* rcmd, struct replay* r);
int replay_load(struct replay_command* rcmd, struct replay* r);
int replay_unload(struct replay_command* rcmd, struct replay* r);
int replay_create_thread(struct replay_command* srcmdp,struct replay *r);
int replay_readpage(struct replay_command* rcmd, struct replay* r);
int replay_readpages(struct replay_command* rcmd, struct replay* r);
int replay_writepage(struct replay_command* rcmd, struct replay* r);
int replay_dget(struct replay_command* srcmdp, struct replay* r);
int replay_dput(struct replay_command* srcmdp, struct replay* r);
int replay_commit_write(struct replay_command* rcmd, struct replay* r);
int replay_sync_page(struct replay_command* srcmdp,struct replay* r);
int replay_mkdir(struct replay_command* rcmd, struct replay* r);
int replay_rmdir(struct replay_command* rcmd, struct replay* r);
int replay_unlink(struct replay_command* rcmd, struct replay* r);
int replay_d_release(struct replay_command* rcmd, struct replay* r);
int replay_readdir(struct replay_command* rcmd, struct replay* r);
int replay_mknod(struct replay_command* rcmd, struct replay* r);
int replay_setattr(struct replay_command* rcmd, struct replay* r);
int replay_fsync(struct replay_command* rcmd, struct replay* r);
int replay_llseek(struct replay_command* rcmd, struct replay* r);
int replay_permission(struct replay_command* rcmd, struct replay* r);

#endif /* _REPLAYK_H */
