/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: tracefs.h,v 1.14 2007/11/04 00:18:26 gopalas Exp $
 */

#ifndef __TRACEFS_H_
#define __TRACEFS_H_

#ifdef __KERNEL__

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define i_size_read(inode) (inode)->i_size
#define i_size_write(inode,size) (inode)->i_size=(size)
#endif

/*
 * STRUCTURES:
 */

#include "common.h"
#include "namecache.h"

/* Turn this on to include code that counts writes and writes in the middle */
/* #undef FIST_COUNT_WRITES */

/* fist file systems superblock magic */
#define TRACEFS_SUPER_MAGIC 0xf15f


/* tracefs inode data in memory */
struct tracefs_inode_info {
    inode_t *wii_inode;
};

/* tracefs dentry data in memory */
struct tracefs_dentry_info {
    dentry_t *wdi_dentry;
};


/* tracefs super-block data in memory */
struct tracefs_sb_info {
    super_block_t *wsi_sb;
    struct vfsmount *hidden_mnt;
    struct trace_sb td;
    struct namecache *namecache;
    char   trace_mmap; /* boolean flag indicating if mmap'ed ops are traced */
	 struct rw_semaphore rwsem;
};

/* file private data. */
struct tracefs_file_info {
    struct file *wfi_file;
};

/*
 * MACROS:
 */
#define tracefs_read_lock(sb)    down_read(&stopd(sb)->rwsem)
#define tracefs_read_unlock(sb)  up_read(&stopd(sb)->rwsem)
#define tracefs_write_lock(sb)   down_write(&stopd(sb)->rwsem)
#define tracefs_write_unlock(sb) up_write(&stopd(sb)->rwsem)


// File TO Private Data
#define ftopd(file) ((struct tracefs_file_info *)((file)->private_data))
#define ftopdlh(file) (((file)->private_data))
// File TO Hidden File
#define ftohf(file) ((ftopd(file))->wfi_file)
// Inode TO Private Data
#define itopd(ino) ((struct tracefs_inode_info *)(ino)->u.generic_ip)
#define itopdlh(ino) ((ino)->u.generic_ip)
// Inode TO Hidden Inode
#define itohi(ino) (itopd(ino)->wii_inode)
#define vnode2lower itohi
// Superblock TO Private Data
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define stopd(super) ((struct tracefs_sb_info *)(super)->u.generic_sbp)
#define stopdlh(super) ((super)->u.generic_sbp)
#else
#define stopd(super) ((struct tracefs_sb_info *)(super)->s_fs_info)
#define stopdlh(super) ((super)->s_fs_info)
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define stoptr(super) (&stopd(super)->td)
#define stopnc(super) (stopd(super)->namecache)
#define mmap_support(super) (stopd(super)->trace_mmap)
#define vfs2priv stopd
// Superblock TO Hidden Superblock
#define stohs(super) (stopd(super)->wsi_sb)
// Dentry TO Hidden Dentry (and private data --- for the future)
#define dtopd(dentry) ((struct tracefs_dentry_info *)(dentry)->d_fsdata)
#define dtopdlh(dentry) ((dentry)->d_fsdata)
#define dtohd(dent) (dtopd(dent)->wdi_dentry)
#define dir2lower dtohd
/* VMArea TO Hidden VMArea */
#define vmatohvma(vma) ((vm_area_t *)((vma)->vm_private_data))

#define sbt(sb) ((sb)->s_type->name)

/* macros to simplify non-SCA code */
#define MALLOC_PAGE_POINTERS(hidden_pages, num_hidden_pages)
#define MALLOC_PAGEDATA_POINTERS(hidden_pages_data, num_hidden_pages)
#define FREE_PAGE_POINTERS(hidden_pages, num)
#define FREE_PAGEDATA_POINTERS(hidden_pages_data, num)
#define FOR_EACH_PAGE
#define CURRENT_HIDDEN_PAGE hidden_page
#define CURRENT_HIDDEN_PAGEDATA hidden_page_data
#define CURRENT_HIDDEN_PAGEINDEX page->index


/*
 * Flags for tracefs_{en,de}code_filename
 * DO_DOTS means the special entries . and .. should be encoded (for symlink)
 * SKIP_DOTS means they should be preserved intact
 */
#define DO_DOTS   0
#define SKIP_DOTS 1

/*
 * EXTERNALS:
 */
extern struct file_operations tracefs_main_fops;
extern struct file_operations tracefs_mmap_fops;
extern struct file_operations tracefs_dir_fops;
extern struct inode_operations tracefs_main_iops;
extern struct inode_operations tracefs_dir_iops;
extern struct inode_operations tracefs_symlink_iops;
extern struct super_operations tracefs_sops;
extern struct dentry_operations tracefs_dops;
extern struct vm_operations_struct tracefs_shared_vmops;
extern struct vm_operations_struct tracefs_private_vmops;
extern struct address_space_operations tracefs_aops;

extern int tracefs_interpose(dentry_t *hidden_dentry, dentry_t *this_dentry, super_block_t *sb, int flag);
#if defined(FIST_FILTER_DATA) || defined(FIST_FILTER_SCA)
extern page_t *tracefs_get1page(file_t *file, int index);
extern int tracefs_fill_zeros(file_t *file, page_t *page, unsigned from);
#endif /* FIST_FILTER_DATA || FIST_FILTER_SCA */

# define tracefs_hidden_dentry(d) __tracefs_hidden_dentry(__FILE__,__FUNCTION__,__LINE__,(d))
extern dentry_t *__tracefs_hidden_dentry(const char *file, const char *func, int line, dentry_t *this_dentry);


extern struct address_space_operations tracefs_aops;
extern page_t *tracefs_get1page(file_t *file, int index);
extern int tracefs_encode_block(const char *from, char *to, int len, const vnode_t *vp, const vfs_t *vfsp, u_long pagenum);
extern int tracefs_decode_block(const char *from, char *to, int len, const vnode_t *vp, const vfs_t *vfsp, u_long pagenum);
extern int tracefs_decode_filename(const char *name, int length, char **decoded_name, int skip_dots, const vnode_t *vp, const vfs_t *vfsp);
extern int tracefs_encode_filename(const char *name, int length, char **encoded_name, int skip_dots, const vnode_t *vp, const vfs_t *vfsp);
#define copy_inode_size(dst, src) \
    dst->i_size = src->i_size; \
    dst->i_blocks = src->i_blocks;

static inline void
fist_copy_attr_atime(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
}
static inline void
fist_copy_attr_ctime(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_ctime = src->i_ctime;
}
static inline void
fist_copy_attr_times(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
}
static inline void
fist_copy_attr_timesizes(inode_t *dest, const inode_t *src)
{
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
    copy_inode_size(dest, src);
}
static inline void
fist_copy_attr_all(inode_t *dest, const inode_t *src)
{
    print_entry_location();
    ASSERT(dest != NULL);
    ASSERT(src != NULL);
    dest->i_mode = src->i_mode;
    dest->i_nlink = src->i_nlink;
    dest->i_uid = src->i_uid;
    dest->i_gid = src->i_gid;
    dest->i_rdev = src->i_rdev;
    dest->i_atime = src->i_atime;
    dest->i_mtime = src->i_mtime;
    dest->i_ctime = src->i_ctime;
    dest->i_blksize = src->i_blksize;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,12)
    dest->i_blkbits = src->i_blkbits;
#endif /* linux 2.4.12 and newer */
    copy_inode_size(dest, src);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    dest->i_attr_flags = src->i_attr_flags;
#endif //i_attr_flags not implemented in Versions >= 2.6
    print_exit_location();
}

/* auto-generated extern definitions */


#endif /* __KERNEL__ */

/*
 * Definitions for user and kernel code
 */

/* ioctls */


#endif	/* not __TRACEFS_H_ */

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
